(********************************************************************
 * Copyright 2011 Jehan Hysseo  (email : jehan at zemarmot.net) {{{
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; version 3 of the License.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA }}}
 *  ******************************************************************)

let block_length = 64;;
let hash_length = 20;;

let word i32 =
    let w = String.make 4 '\000' in
    let rec aux = function
    | z, _ when z = Int32.zero ->
        (* This will occure at latest just after octet 0 *)
        w
    | div, i ->
        let rem = Int32.to_int (Int32.logand div 0xFFl) in
        w.[i] <- Char.chr rem;
        aux (Int32.shift_right_logical div 8, i - 1)
    in aux (i32, 3)
;;

let word_to_int32 word =
    let rec aux value = function
        | 3 ->
            let c =
                Int32.shift_left (Int32.of_int (Char.code word.[3])) 0
            in Int32.add value c
        | i ->
            let c =
                Int32.shift_left (Int32.of_int (Char.code word.[i])) (8 * (3 - i))
            in aux (Int32.add value c) (i + 1)
    in aux Int32.zero 0
;;

(* XXX: this function expects a positive t. As this is internal function and I
* look for optimization (but does it really help a single test less?), I don't
* test for negative t. *)
let k = function
    (*| t when t < 0 -> failwith "Sha1.k: logic error: negative t."*)
    | t when t < 20 ->
        0x5A827999l
    | t when t < 40 ->
        0x6ED9EBA1l
    | t when t < 60 ->
        0x8F1BBCDCl
    | t when t < 80 ->
        0xCA62C1D6l
    | _ -> failwith "Sha1.k: logic error, t too big."
;;

(* XXX: I don't check the negative t. *)
let f b c d = function
    (*| t when t < 0 -> failwith "Sha1.f: logic error: negative t."*)
    | t when t < 20 ->
        Int32.logor (Int32.logand b c) (Int32.logand (Int32.lognot b) d)
    | t when t < 40 ->
        Int32.logxor (Int32.logxor b c) d
    | t when t < 60 ->
        Int32.logor (Int32.logor (Int32.logand b c)
            (Int32.logand b d)) (Int32.logand c d)
    | t when t < 80 ->
        Int32.logxor (Int32.logxor b c) d
    | _ -> failwith "Sha1.f: logic error, t too big."
;;

(***********)
(* PADDING *)
(***********)

(* The length of a message is the number of bit in the message.
* As we expect it to be less than 2^64, we use Int32. *)
let message_length message =
    let l = Int64.of_int (String.length message) in (*length in octet *)
    Int64.mul l (8L)
;;

(* Produce a 64 bits string for the given length.
* I expect l to be less than 2^64. *)
let length_to_string l =
    let s = String.make 8 '\000' in
    let rec aux i = function
        | z when z = Int64.zero -> s
        | div ->
            begin
                (* XXX: equivalent to: Int64.rem div 256L **)
                match Int64.logand div 255L with
                | z when z = Int64.zero ->
                    (* XXX: equivalent to: Int64.div div 256L **)
                    aux (i - 1) (Int64.shift_right_logical div 8)
                | r ->
                    let c = Char.chr (Int64.to_int r) in
                    s.[i] <- c;
                    aux (i - 1) (Int64.shift_right_logical div 8)
            end
    in aux 7 l
;;

(* A padded message is made of blocks of 512 bits long (64 octets).
* A one bit with zeros needs to be inserted, so at least one Int32.additional octet
* will be necessary at the end.
* 64 bits (8 octets) are reserved at the end for length of whole message.
* Hence if the last block is more than 64 - 8 - 1 = 55 octets, I fill it with
* zeros and create a new block.
* @return a list of 512 bits string blocks.*)
let pad message =
    let length_pad = length_to_string (message_length message) in
    let rec aux blocks remains =
        let size = String.length remains in
        match size with
        | s when s >= 64 ->
            aux ((String.sub remains 0 64)::blocks) (String.sub remains 64 (s - 64))
        | s when s > 55 ->
            let pad_begin = String.make (64 - s) '\000' in
            pad_begin.[0] <- '\x80';
            let pad_end = String.make 56 '\000' in
            (pad_end ^ length_pad)::(remains ^ pad_begin)::blocks
        | s ->
            let pad = String.make (56 - s) '\000' in
            pad.[0] <- '\x80';
            (remains ^ pad ^ length_pad)::blocks
    in List.rev (aux [] message)
;;

(* These are the initial values of h0-4. *)
let h = function
    | 0 ->
        0x67452301l
    | 1 ->
        0xEFCDAB89l
    | 2 ->
        0x98BADCFEl
    | 3 ->
        0x10325476l
    | 4 ->
        0xC3D2E1F0l
    | _ -> failwith "Sha1.h: unexpected error."
;;

let circ_left_shift w n =
    let w1 = Int32.shift_left w n in
    let w2 = Int32.shift_right_logical w (32 - n) in
    Int32.logor w1 w2
    (* XXX: considering that the bits are disjoint, this is equivalent to an
* Int32.add. Though after making some tests to check which is faster, I did not see
* much difference... *)
;;

(* mi is supposed to be 64 octets long. I won't check this again in this private
* function. *)
let make_w mi =
    let w = Array.make 80 Int32.zero in
    let rec aux = function
        | 80 -> w
        | i when i < 16 ->
            w.(i) <- word_to_int32 (String.sub mi (i * 4) 4);
            aux (i + 1)
        | t ->
            w.(t) <- circ_left_shift
                (Int32.logxor (Int32.logxor (Int32.logxor w.(t - 3) w.(t - 8)) w.(t - 14)) w.(t - 16)) 1;
            aux (t + 1)
    in aux 0
;;
        
let digest message =
    let m = pad message in
    let rec aux h0 h1 h2 h3 h4 = function
    | [] ->
        (word h0) ^ (word h1) ^
        (word h2) ^ (word h3) ^
        (word h4)
    | mi::l ->
        let w = make_w mi in
        let rec loop a b c d e = function
        | 80 -> a, b, c, d, e
        | t ->
            let temp =
                Int32.add
                    (Int32.add (circ_left_shift a 5) (f b c d t))
                    (Int32.add (Int32.add e w.(t)) (k t)) in
            loop temp a (circ_left_shift b 30) c d (t + 1)
        in
        let a, b, c, d, e = loop h0 h1 h2 h3 h4 0 in
        aux (Int32.add h0 a) (Int32.add h1 b) (Int32.add h2 c)
            (Int32.add h3 d) (Int32.add h4 e) l
    in aux (h 0) (h 1) (h 2) (h 3) (h 4) m
;;
